use std::{io, sync::Arc};

use rocket::{
    fairing::{Fairing, Info, Kind},
    fs::NamedFile,
    http::Header,
    Request,
    Responder,
    Response,
};

/// Value of the cache control header
pub static CACHE_HEADER_VALUE: &str = "max-age=604800, must-revalidate";

/// Fairing for adding cache control headers to certain routes
///
/// ## Example
/// ```rust
/// # use cache_control::CacheControl;
/// # use rocket::{
/// #     fs::FileServer,
/// #     Build,
/// #     Rocket,
/// # };
///
/// pub fn init(build: Rocket<Build>) -> Rocket<Build> {
///     build
///         .mount("/img", FileServer::from("./img"))
///         .mount("/js", FileServer::from("./js"))
///         .attach(CacheControl::new(&["/img", "/js"]))
/// }
/// ```
#[derive(Debug)]
pub struct CacheControl {
    routes: Arc<&'static [&'static str]>,
}

impl CacheControl {
    /// Specify the route bases you want to be cached
    pub fn new(routes: &'static [&'static str]) -> Self {
        Self {
            routes: Arc::new(routes),
        }
    }
}

#[rocket::async_trait]
impl Fairing for CacheControl {
    fn info(&self) -> Info {
        Info {
            name: "Cache Control",
            kind: Kind::Response,
        }
    }

    async fn on_response<'r>(&self, req: &'r Request<'_>, res: &mut Response<'r>) {
        if let Some(route) = req.route() {
            let uri = route.uri.base.path().as_str();

            if self.routes.iter().any(|r| *r == uri) {
                let header = Header::new("Cache-Control", CACHE_HEADER_VALUE);
                res.set_header(header);

                log::info!("Setting cache headers on {uri}");
            }
        }
    }
}

/// Responder for a cached file, useful when serving named files directly
#[derive(Responder)]
pub struct CachedFile {
    inner:  io::Result<NamedFile>,
    header: Header<'static>,
}

impl CachedFile {
    /// Respond with a file and append the cache control headers
    pub fn respond(file: io::Result<NamedFile>) -> Self {
        CachedFile {
            inner:  file,
            header: Header::new("Cache-Control", CACHE_HEADER_VALUE),
        }
    }
}
